// SHT1x Humidity & Temperature Sensor

#include <16F873.h>

//#device ICD=TRUE

// Pointer size = 16bit
#device adc=10 *=16

// Clock = 2MHz
#use delay(clock=2000000)

// Configuration Fuses
#fuses XT, WDT, NOWRT,NOPROTECT, NOPUT, NOBROWNOUT, NOLVP, NOCPD

// LCD Control & Data signals line

// RB2 : RS (Register select)
// RB1 : R/W
// RB0 : E (Enable signal)
//
// RC0 : DB0
// RC1 : DB1
// RC2 : DB2
// RC3 : DB3
// RC4 : DB4
// RC5 : DB5
// RC6 : DB6
// RC7 : DB7

// LCD Control Signals
#define RS_LOW  0b000
#define RS_HIGH 0b100
#define RW_R    0b10
#define RW_W    0b00
#define E_LOW   0b0
#define E_HIGH  0b1

#use fast_io(A)

// Delay Time
#define DLTIMEu 10

// LCD char buf
char  lbuf[21];

// CRC
int8  crcdata;


int8 const crc_table[256]= {
0, 49, 98, 83, 196, 245,166, 151,185, 136,219, 234,125, 76, 31, 46, 67, 114,33, 16,
135, 182,229, 212,250, 203,152, 169,62, 15, 92, 109,134, 183,228, 213,66, 115,32, 17,
63, 14, 93, 108,251, 202,153, 168,197, 244,167, 150,1, 48, 99, 82, 124, 77, 30, 47,
184, 137,218, 235,61, 12, 95, 110,249, 200,155, 170,132, 181,230, 215,64, 113,34, 19,
126, 79, 28, 45, 186, 139,216, 233,199, 246,165, 148,3, 50, 97, 80, 187, 138,217, 232,
127, 78, 29, 44, 2, 51, 96, 81, 198, 247,164, 149,248, 201,154, 171,60, 13, 94, 111,
65, 112,35, 18, 133, 180,231, 214,122, 75, 24, 41, 190, 143,220, 237,195, 242,161, 144,
7, 54, 101, 84, 57, 8, 91, 106,253, 204,159, 174,128, 177,226, 211,68, 117,38, 23,
252, 205,158, 175,56, 9, 90, 107,69, 116,39, 22, 129, 176,227, 210,191, 142,221, 236,
123, 74, 25, 40, 6, 55, 100, 85, 194, 243,160, 145,71, 118,37, 20, 131, 178,225, 208,
254, 207,156, 173,58, 11, 88, 105,4, 53, 102, 87, 192, 241,162, 147,189, 140,223, 238,
121, 72, 27, 42, 193, 240,163, 146,5, 52, 103, 86, 120, 73, 26, 43, 188, 141,222, 239,
130, 179,224, 209,70, 119,36, 21, 59, 10, 89, 104,255, 206,157, 172};



// SHT1x Control signal

// *******************************
// DATA Line = Low, SCK Line = Low
// *******************************
void DL_SL()
{
   output_a(0b001111);
   delay_us(DLTIMEu);
}

// *******************************
// DATA Line = High, SCK Line = Low
// *******************************
void DH_SL()
{
   output_a(0b011111);
   delay_us(DLTIMEu);
}

// *******************************
// DATA Line = Low, SCK Line = High
// *******************************
void DL_SH()
{
   output_a(0b101111);
   delay_us(DLTIMEu);
}

// *******************************
// DATA Line = High, SCK Line = High
// *******************************
void DH_SH()
{
   output_a(0b111111);
   delay_us(DLTIMEu);
}

// *******************************
// "Transmission Start" sequence
// *******************************
void  SHT_Trans_Start()
{
   // PORTA initial value
   output_a(0xFF);

   SET_TRIS_A( 0b000000 ); // RA4 = output

   DH_SL();
   DH_SH();
   DL_SH();
   DL_SL();
   DL_SH();
   DH_SH();
   DH_SL();

   // Wait
   delay_ms(10);

}

// *******************************
// LCD RS Line = Low, DATA Wite
// *******************************
void LCD_RSLOW_RW(int lcd_data)
{
   int8 i;

   i = lcd_data;
   output_c(i);

   i = RS_LOW + RW_W + E_LOW;
   output_b(i);
   delay_us(DLTIMEu);

   i = RS_LOW + RW_W + E_HIGH;
   output_b(i);
   delay_us(DLTIMEu);

   i = RS_LOW + RW_W + E_LOW;
   output_b(i);
   delay_us(DLTIMEu);

}

// *******************************
// LCD RS Line = High, DATA Wite
// *******************************
void LCD_RSHI_RW(int lcd_data)
{
   int8 i;

   i = lcd_data;
   output_c(i);

   i = RS_HIGH + RW_W + E_LOW;
   output_b(i);
   delay_us(DLTIMEu);

   i = RS_HIGH + RW_W + E_HIGH;
   output_b(i);
   delay_us(DLTIMEu);

   i = RS_HIGH + RW_W + E_LOW;
   output_b(i);
   delay_us(DLTIMEu);

}

// *******************************
// LCD Data output
// *******************************
void   LCD_Dataout(int8 i)
{
   i += 0b10000000;
   LCD_RSLOW_RW(i);
   i = 0;
   while(lbuf[i] != 0) {
      LCD_RSHI_RW(lbuf[i]);
      i ++;
   }
}

// *******************************
// SHT Command transmission
// *******************************
void SHT_Trans(int8 data)
{
   int8 i;

// crc
   crcdata = 0;
   crcdata ^= data;
   crcdata = crc_table[crcdata];

   // Command transmission
   DH_SL();
   for(i = 0 ; i < 8 ; i++) {
      if((data & 0b10000000) == 0) {
         DL_SL();
         DL_SH();
         DL_SL();
      } else {
         DH_SL();
         DH_SH();
         DH_SL();
      }
      data <<= 1;
   }

   // The ACK signal is received.
   SET_TRIS_A( 0b010000 ); // RA4 = input
   DH_SL();
   DH_SH();
   if (input(PIN_A4) == 1) {
      // Display Clear
      LCD_RSLOW_RW(0b00000001);
      // Cursor Home
      LCD_RSLOW_RW(0b00000010);
      delay_ms(10);
      // Error Message
      sprintf(lbuf,"ERROR;");
      LCD_Dataout(0x00);
      sprintf(lbuf,"SHT does not respond");
      LCD_Dataout(0x40);
      sprintf(lbuf," to the command.");
      LCD_Dataout(0x14);

      restart_wdt();
      sleep();
      reset_cpu();
   }
   DH_SL();
}


// *******************************
// SHT data recived
// *******************************
int16   SHT_data_recive()
{
   int8 rdata1,rdata2;
   int16 rdata;
   int8 i;
   int8  read_crc;

   // MSB 8bit of the temperature data is received.
   rdata1 = 0;
   for(i = 0 ; i < 8 ; i++) {
      DH_SH();
      rdata1 <<= 1;
      if(input(PIN_A4) == 1) rdata1 ++;
      DH_SL();
   }

   crcdata ^= rdata1;
   crcdata = crc_table[crcdata];

   // The ACK signal is transmitted.
   SET_TRIS_A( 0b000000 ); // RA4 = output

   // ACK Puls
   DL_SL();
   DL_SH();
   DL_SL();

   SET_TRIS_A( 0b010000 ); // RA4 = input

   // LSB 8bit of the temperature data is received.
   rdata2 = 0;
   for(i = 0 ; i < 8 ; i++) {
      DH_SH();
      rdata2 <<= 1;
      if(input(PIN_A4) == 1) rdata2 ++;
      DH_SL();
   }
   crcdata ^= rdata2;
   crcdata = crc_table[crcdata];

   SET_TRIS_A( 0b000000 ); // RA4 = output

   // ACK Puls(CRC used)
   DL_SL();
   DL_SH();
   DL_SL();

   // The CRC signal is received.
   SET_TRIS_A( 0b010000 ); // RA4 = input

   read_crc = 0;
   for(i = 0 ; i < 8 ; i ++) {
      DH_SH();
      read_crc >>= 1;
      if(input(PIN_A4) == 1) read_crc += 0b10000000;
      DH_SL();
   }
   DH_SL();

   // ACK Puls
   DL_SL();
   DH_SH();
   DH_SL();

   // CRC check
   if (crcdata != read_crc) {
      // Display Clear
      LCD_RSLOW_RW(0b00000001);
      // Cursor Home
      LCD_RSLOW_RW(0b00000010);
      delay_ms(10);
      // Error Message
      sprintf(lbuf,"ERROR;");
      LCD_Dataout(0x00);
      sprintf(lbuf,"CRC Error!");
      LCD_Dataout(0x40);
      sprintf(lbuf,"Calc CRC : %02X",crcdata);
      LCD_Dataout(0x14);
      sprintf(lbuf,"Read CRC : %02X",read_crc);
      LCD_Dataout(0x54);

      restart_wdt();
      sleep();
      reset_cpu();
   }

   // MSB and LSB of the temperature data are stored in 16bit variable.
   rdata = (int16)rdata1*256 + (int16)rdata2;

   // Wait
   delay_ms(100);

   return(rdata);
}

// *******************************
// main function
// *******************************
void main()
{
   int8 i;
   int8  sleep_count;

   float tval;
   float rhval;
   float x;
   int16 rdata;

   float tv_max;  // maximum Temperature
   float tv_min;  // minimum Temperature
   float rh_max;  // maximum Humidity
   float rh_min;  // maximum Humidity

   // ADC is not used.
   setup_adc_ports(NO_ANALOGS);
   setup_adc(ADC_OFF);

   // Port initial value
   output_a(0xFF);
   output_b(0xFF);
   output_c(0xFF);

   // All PORTB and PORTC are output.
   SET_TRIS_C( 0b00000000);
   SET_TRIS_B( 0b00000000);

   // Setting of PORTA
   SET_TRIS_A( 0b000000 );

   // Powter UP , Variable initialization
   if (restart_cause() == NORMAL_POWER_UP) {
      tv_max = -100.0;
      tv_min = 200.0;
      rh_max = 0;
      rh_min = 100;
      sleep_count = 0;

      // *******************************
      // LCD module initialization
      // *******************************

      delay_ms(50);

      LCD_RSLOW_RW(0b00110000);

      delay_ms(5);

      LCD_RSLOW_RW(0b00110000);

      delay_us(100);

      LCD_RSLOW_RW(0b00110000);

      // 8bit,1/16 duty
      LCD_RSLOW_RW(0b00111000);

      // Display ON(bit2 = 1),Cursor OFF(bit1 = 0),NO Blink(bit0 = 0)
      LCD_RSLOW_RW(0b00001100);

      // Display Clear
      LCD_RSLOW_RW(0b00000001);

      // Cursor Home
      LCD_RSLOW_RW(0b00000010);

   }

   // WDT is set in the maximum value.
   setup_wdt(WDT_2304MS);

   // Sleep count down & sleep_count check

   if (sleep_count != 0) {
      sleep_count --;
      sleep();
      reset_cpu();
   }
   sleep_count = 4;

   // "Transmission Start" sequence
   SHT_Trans_Start();

   // *******************************
   // Temperature measurement command
   // *******************************
   SHT_Trans(0b00000011); // Command transmission


   // SHT Time to measure temperature by 14bit accuracy
   // 210ms * 1.15 = 315ms
   delay_ms(210);
   
   // End Conversion check
   while(input(PIN_A4) == 1);

   // SHT data recived
   rdata = SHT_data_recive();

   // Temperature = d1 + d2 x rdata
   // d1 = -40 (VDD = 5V), d2 = 0.01 (14bit)
   tval = (float)rdata * 0.01 - 40.0;

   // "Transmission Start" sequence
   SHT_Trans_Start();

   // *******************************
   // Humidity measurement command
   // *******************************
   SHT_Trans(0b00000101); // Command transmission

   // Time to measure temperature by 12bit accuracy
   // 55ms
   delay_ms(210);
   
   // End Conversion check
   while(input(PIN_A4) == 1);

   // SHT data recived
   rdata = SHT_data_recive();

   // RH = C1 + C2 x rdata + C3 x rdata x rdata
   x = (float)rdata;
   rhval = -4 + 0.0405 * x - 0.0000028 * x * x;

   // *******************************
   // DATA output
   // *******************************

   // Temperature output
   i = 0xDF;
   sprintf(lbuf,"Tmp = %5.1f (%CC)",tval,i);

   // Data output
   LCD_Dataout(0);

   // Memory of the maximum temperature and minimum temperature
   if (tval > tv_max ) tv_max = tval;
   if (tval < tv_min ) tv_min = tval;

   // The maximum temperature and a minimum temperature are
   // written in the second line.

   sprintf(lbuf," max=%5.1f,min=%5.1f",tv_max,tv_min);
   // Data output
   LCD_Dataout(0x40);


   // Humidity is written in the third line.
   sprintf(lbuf,"RH  = %5.1f (%%)",rhval);

   // Data output
   LCD_Dataout(0x14);

   // Memory of the maximum Humidity and minimum Humidity
   if (rhval > rh_max ) rh_max = rhval;
   if (rhval < rh_min ) rh_min = rhval;

   // The maximum Humidity and a minimum Humidity are
   // written in the second line.
   sprintf(lbuf," max=%5.1f,min=%5.1f",rh_max,rh_min);

   // Data output
   LCD_Dataout(0x54);

// Sleep
   restart_wdt();
   sleep();
   reset_cpu();
}
